/*	control.c

	AFAPI control function definitions and data for TTL_PAPERS.

	July 1996 by Hank Dietz
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>


#include "afapi.h"
#include "inline.h"


int iproc = -1;			/* PE number (0..NPROC-1) */
int cproc = -1;			/* Console PE number */

volatile p_shm_type *p_shm_all = 0;	/* Pointer to all PE info */
volatile p_shm_type *p_shm_me = 0;	/* Pointer to my PE info */

int p_shm_id = -1;			/* System V shared segment ID */


static int p_init_state = 0;	/* 0=never, 1=previously, 2=now */

/*	p_init...

	This routine is called to check-in with SHMAPERS.  It must be
	called before any other p_ or s_ routines.
*/

int
p_init(void)
{
	/* Initialize the SHMAPERS variables and create the System V
	   shared memory segment....
	*/
	register char *p;


	/* Step 0:  Have we done this before? */
	switch (p_init_state) {
	default:
		fprintf(stderr,
"AFAPI PE:  in p_init(), p_init_state=%d can't happen?\n",
			p_init_state);
		return(1);
	case 2:
		/* Call to init when we're already init?
		   We'll allow it, but do a p_confirm().
		*/
		goto step6;
	case 1:
		/* Yup, just need a partial check-in */
		goto step6;
	case 0:
		break;
	}


	/* Step 1:  Examine environment variables */
	p = getenv("AFAPI");
	if ((p == NULL) || strcmp(AFAPI, p)) {
		fprintf(stderr,
"AFAPI PE:  need %s hardware (set AFAPI environment variable)\n",
			AFAPI);
		return(2);
	}

	p = getenv("AFAPI_VERSION");
	if (p == NULL) p = "not given";
	if (strcmp(AFAPI_VERSION, p)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_VERSION of code is %s, environment is %s\n",
			AFAPI_VERSION, p);
		return(2);
	}

	p = getenv("AFAPI_NPROC");
	if ((p != NULL) && (NPROC != atoi(p))) {
		fprintf(stderr,
"AFAPI PE:  compiled for AFAPI_NPROC=%d, environment has %d\n",
			NPROC, atoi(p));
		return(2);
	}

	/* Ignore AFAPI_IPROC...  we'll set those */

	p = getenv("AFAPI_CPROC");
	cproc = ((p == NULL) ? -1 : atoi(p));
	if ((cproc < 0) || (cproc >= NPROC)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_CPROC is not a legal value (from 0 to %d)\n",
			(NPROC - 1));
		return(2);
	}


	/* Step 2:  Create the System V shared memory segment.

	   There are quite a few options for this.  One issue is if we
	   want the shared memory segment to be visible to other
	   programs or if we want it strictly local; this makes it
	   local to avoid security problems.  Another issue is how to
	   set the segment access permission; again, this keeps it for
	   us only.  There also are options for how to destroy the
	   segment when we are done; this code sets it so that the
	   segment will automatically disappear when the last PE
	   process terminates (otherwise, abnormal terminations could
	   result in a slow virtual memory leak).

	   Notice that the segment is created one struct oversize; the
	   NPROC struct is used for broadcast and the extra space
	   holds broadcast block data.
	*/
	p_shm_id = shmget(IPC_PRIVATE,
			  (((NPROC + 1) * sizeof(p_shm_type)) +
			   (2 * P_BLOCKSIZE)),
			  (IPC_CREAT | 0600));
	if (p_shm_id < 0) {
		fprintf(stderr,
"AFAPI PE%d:  can't get shared memory segment (is System V IPC enabled?)\n",
			IPROC);
		return(3);
	}
	p_shm_all = ((p_shm_type *) shmat(p_shm_id,
					  0,	/* put it anywhere */
					  0));	/* not read only */
	if (((int) p_shm_all) == -1) {
		/* Why did AT&T make -1 the error value for this
		   routine that is supposed to return a pointer?
		*/
		fprintf(stderr,
"AFAPI PE%d:  can't attach to shared memory segment %d\n",
			IPROC, p_shm_id);
		return(3);
	}
	if (shmctl(p_shm_id, IPC_RMID, 0) != 0) {
		fprintf(stderr,
"AFAPI PE%d:  can't set shared memory segment %d to self-destruct\n",
			IPROC, p_shm_id);
		shmdt((char *) p_shm_all);	/* destroy it */
		return(3);
	}
		

	/* Step 3:  Initialize the shared memory information.

	   It is important that this be done before we have NPROC
	   processes around....  Also, gotta love those "register
	   volatile" pointer declarations....  ;-)
	*/
	{
		register int i;
		register volatile p_shm_type *p = p_shm_all;

		for (i=0; i<NPROC; ++i) {
			p->datum[0].u_64u = 0;
			p->datum[1].u_64u = 0;
			p->barno = 0;
			p->pid = 0;
			++p;
		}
	}


	/* Step 4:  Time to fork() the PE processes....

	   This is an ordinary fork(), which clones the connection to
	   the shared memory segment, but otherwise gives each PE its
	   own unique copy of the writeable portion of the address
	   space.  This is important because we don't have to worry
	   about system libraries being thread safe, etc.
	*/
	for (iproc=NPROC-1; iproc>0; --iproc) {
		if (fork() == 0) break;
	}


	/* Step 5:  Wake-up the PEs and do an initial barrier.

	   Close the stdin, stdout, and stderr connections of everyone
	   but CPROC.  Then set the shared memory pointer for each
	   PE.  Finally, signal that you are there by filling-in your
	   PID and waiting to see that everyone else has also done so.
	*/
	if (IPROC != CPROC) {
		/* We are not the console, so disconnect us from
		   stdin, stdout, stderr...  then connect everything
		   to /dev/null to ensure no harm.
		*/
		close(0);
		close(1);
		close(2);
		open("/dev/null", O_RDONLY);
		open("/dev/null", O_WRONLY);
		open("/dev/null", O_WRONLY);
	}
	p_shm_me = (p_shm_all + IPROC);
	p_shm_me->pid = getpid();
	{
		/* And now the initial mutatant barrier... */
		register int i;
		register volatile p_shm_type *p = p_shm_all;

		for (i=0; i<NPROC; ++i) {
			while (p->pid == 0) ;
			++p;
		}
	}
	/* And finally a regular barrier... */
	p__wait();


	/* Step 6:  Do a p_confirm(). */
step6:
	if (p_confirm() != 0) {
		fprintf(stderr,
"AFAPI PE%d:  p_confirm() suggests faulty hardware or connection\n",
			IPROC);
		return(7);
	}


	/* Step 7:  We're in! */
	p_init_state = 2;
	return(0);
}

/*	p_exit()...

	Do not need to do anything for this, because the shared memory
	segment will go away by itself when the last PE process
	terminates....
*/

void
p_exit(void)
{
	switch (p_init_state) {
	default:
		fprintf(stderr,
"AFAPI PE:  in p_exit(), p_init_state=%d can't happen?\n",
			p_init_state);
		exit(1);
	case 0:
	case 1:
		/* We aren't even checked in...  so maybe the luser
		   wants us to confirm that we are really checked
		   out?  Paranoia?  Well, ok.  Why argue?
		*/
		break;
	case 2:
		/* Ok, let's say we are checked out... */
		p_wait();
		p_init_state = 1;
	}
}

/*	p_confirm()...

	Attempt to confirm the previous operations.  Well, a hardware
	malfunction in an SMP is not likely to leave this code alive,
	so all we need to check for is potential loss of
	synchronization due to a software bug.  In any case, one can
	safely assume that correct operation has resumed by the time
	this code returns.
*/

int
p_confirm(void)
{
	register int errs = 0;
	register int i;

again:
	while (p_reduceOr32(8) != 8) ++errs;
	if (p_reduceOr32(2) != 2) goto again;
	if (p_reduceOr32(4) != 4) goto again;
	if (p_reduceOr32(1) != 1) goto again;

	for (i=0; i<NPROC; ++i) {
		register p_bar mask = (1 << i);

		if (p_reduceOrbar((i == IPROC) ? mask : 0) != mask) {
			goto again;
		}
	}

	/* Return 1 for error, 0 otherwise */
	return(p_reduceOr32(errs != 0));
}
